% Richards equation parameter inverse problem
% clean driver for inverse porous media flow
% Recovers C and K from C(u_)*u_t = (K(u)*(u_x-1))_x
% where g (u at x=0) and h (K(u)*(dudx-1) at x=1) are measured data 
% integral identity is used to recover C(u) and K(u) 
% matlab solvers used to integrate in time

close all;
clear;
clear diag_sys;
clear coeff_C;
clear coeff_K;

%%%%%%%%%%%%%%%%%%%%%
%%%% GLOBAL VARS %%%% 
global XXX TTT

%%%%%%%%%%%%%%%%%%%%%%%%
%%% Parameter  are read in %%
%%% Edit paramter file in order to change these values

PARAMETER_FILE = 'parameter.txt';

readit(PARAMETER_FILE);

Win_x = max(2,ceil(Win_x)); 
Win_t = max(2,ceil(Win_t));

% END parameter section
%%%%%%%%%%%%%%%%%%%%%%

% BUILD  grids 
x = linspace(Xmin,Xmax,Xstep);       % x grid
t = linspace(Tmin,Tmax,Tstep);       % time grid
XXX = x;                             % Triple Caps are global variables
TTT = t;
dt = (Tmax - Tmin) / Tstep;
if sum(Win_t)<Tstep;
   Win_t(end+1) = Tstep - sum(Win_t);
end

Min_w = Win_t * dt 

keyboard
%%%%%%%%%%%%%%%%%%
% EXPERIMENT

% Run EXPERIMENT to get boundary data
% Solves Ctrue*u_t = (Ktrue*u_x)_x
% with bc's u_x(0,t) = BCs(:,2)(t), u(1,t) = BC_data(:,3)(t)
% with Ctrue Ktrue known functions 
% and ic u(x,0) = 0;

if experiment == 1
    BCs = [Tmin 1  0  ;
           Tmax 1  -3 ];
%    u0 = 0*x(2:end-1);
    u0 = x(2:end-1)-1;  %steady state soln  
    tic
    [te,gtrue,htrue,utrue] = exper(Solver1,Udot,t,u0,Ctrue,Ktrue,BCs);
    toc
    ge = perturb(gtrue,G_err,G_eps);
    he = perturb(htrue,H_err,H_eps);
        
 %%%%%%%%%%%%%%%%%%%%%
        
    C0 = evaluate(Ctrue,u0(1));
    K0 = evaluate(Ktrue,u0(1));
        
    fprintf('\n');
    fprintf('Done making the the experimental data');
    fprintf('\n');
else
    data = load('data.txt');
    te = data(:,1);
    ge = data(:,2);
    he = data(:,3);
    K0 = 1;
    C0 = .5;
end
ex_data = [ te , ge , he ];

%break;

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% COEFFICIENT ID  
%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Now we attempt to ID coefficient C(u) and K(u)
% We solve direct (forward in time) and produce u_x
% and dual (back in time) to get phi_x  
% on Q_t (Q_level) strips
% Forward data will come from experiment data ge and he
% Dual data (theta) will be either ge, he or a constant function
% there is one dual problem 
% dualdot.m 
% each with different boundary values 

% Sets up initial approximation of coefficient
% Here I need two points for linear interp

minu = min(ge);
maxu = max(ge);

init_C = [ maxu , C0 ];     %  This is presumed to be known
init_K = [ maxu , K0 ];     %  This is presumed to be known

app_C = [  init_C ;
    minu , C0 ] ;     % approx_C is the form [ u , D(u) ]
app_K = [  init_K ;
    minu , K0 ] ;     % approx_K is the form [ u , D(u) ]

% initialize to get loop started

drel = [];
tlevel = [];
flagIt = 0;
flagC = 0;
flagK = 0;

tf = 0;
iter = 1;
full_tspan = [];

while( tf<Tmax & iter<Tstep )
    
    if norm(drel,inf)>Tol & flagIt<Maxit
        iter = iter - 1;
        flagIt = flagIt + 1;
        fprintf('Recomputing %d iterate of strip # %d \n,',flagIt,iter);
        fprintf('since relative tolerance '); 
        fprintf(' (|g|/max(g) + |h|/max(h))= %3.2d > %3.2d \n',max(drel),Tol)
    end
    
    % Set up initial conditions
    u0 = 0*x(2:end-1);
    ts = t(1);
    tf = Tmax;
    
    u_full = [NaN , NaN , x];
    data_full = NaN.*ones(1,4);
    ud_full = [];
    phi1_full = [];
    phi2_full = [];
    td_full = [];
    ux_full = [];
    ut_full = [];
    phi1x_full = [];
    phi2x_full = [];
    phi1t_full = [];
    phi2t_full = [];
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%55
    
    M = [];
    b = [];
    dCK = [];
    dC = [];
    dK = [];
    
    level = 1;
    flag = 0;
    while (level<=iter & flag ~= 1 )
        %    while ((iter-1)<=iter & flag ~= 1 )
        %        fprintf('\nSolving forward problem on strip %i of %i. \n',level,iter);
        tspan_0 = [ts,Tmax];
        uda = [];
        [dsol{level},tspan,flag] = solve_forward(Solver2,Udot,tspan_0,u0...
            ,app_C,app_K,BCs,Win_x,Min_w(level));
        tf = max(tspan);
        
        fprintf('Time integration stopped at time %0.2g of %g\n',tf,Tmax);
        %        fprintf('Finished solve on level %g \n',level);
        
        if flag == 1, 
            tspan = [ts,Tmax];
            fprintf('Warning: Last region is small, so results may be inaccurate.\n');
        end
        
        [ex_nodes,tspan_common] = find_nodes(ex_data(:,1),tspan);
        ttt_common{level} = tspan_common;
        if isempty(tspan_common) == 0
            [tda,uda,gda,hda] = make_data(dsol{level},tspan_common,app_K,BCs);
            deltag = ex_data(ex_nodes,2) - gda;
            dgrel = abs(deltag ./ max(abs(ex_data(:,2))));
            deltah = ex_data(ex_nodes,3) - hda;
            dhrel = abs(deltah ./ max(abs(ex_data(:,3))));
            drel = dgrel + dhrel;
            %           fprintf('\nThis is using only Ku h data \n');
            % I could use Ku h data as well
        else
            deltag = 0; 
            dgrel = 0;
            deltah = 0; 
            dhrel = 0;
            drel = dgrel + dhrel;
        end
        UDA{level} = uda;
        ddg{level} = deltag;
        ddh{level} = deltah;
        tlevel = tspan(2);
        ttspan{level} = tspan;
        full_tspan = union(tspan,full_tspan);
        %	fprintf('making dual data... \n');
        % stored in tau coordinantes (flips forward data in time)
        data1{level} = [tspan_common , ...
                average(deltag(end:-1:1),G_const,Zeta1),  ...
                0 * average(deltah(end:-1:1),H_const,Zeta2) ];
        data2{level} = [tspan_common , ... 
                0 * average(deltag(end:-1:1),G_const,Zeta1),  ...
                average(deltah(end:-1:1),H_const,Zeta2) ];
%        data2{level} = [tspan_common , ...
%                 0 * average(deltag,G_const,Zeta1),  ...
%           ((tspan_common - min(tspan_common))./max(tspan_common)-1) ];
        % Remove repeats in data QUICK FIX
        [C,IA,IB] = union(uda(:,1),uda(:,1));
        SIA = sort(IB);
        
        %        this was changed data1{level} <- dual_data{level}.one
        %        this was changed data2{level} <- dual_data{level}.two
        
        dual_data{level}.one = data1{level}(SIA,:);
        dual_data{level}.two = data2{level}(SIA,:);
        
        if flag ~= 1 
            ts = dsol{level}.xe;  %reset time to start at end of previous strip
            u0 = dsol{level}.ye;   %set initial condition to soln at time ts
        end
        level = level+1;
    end
    level = level-1;
    
    phi0.one = 0*x(2:end-1);
    phi0.two = 0*x(2:end-1);
    dlevel = level;
    while (dlevel >= 1) 
        fprintf('Solving dual on strip %d \n',dlevel);
        avgu = mean(UDA{dlevel},1); 
        Chat = evaluate(app_C,avgu);
        Khat = evaluate(app_K,avgu);
        dual_app_K = [avgu(:) , Khat(:)];
        Khatp = [ dual_app_K(1:end-1,1) , diff(dual_app_K(:,2)) ...
                ./ diff(dual_app_K(:,1))];
        dualsol{dlevel} = solve_dual(Solver2,Phidot,ttspan{dlevel},phi0,Chat,Khat,Khatp,dual_data{dlevel}); 
        
        %     Start to construct matrix entries for current timespan
        disp('build matrix in current tspan');
        [td,ud,gd,hd] = make_data(dsol{dlevel},ttspan{dlevel},app_K,...
            BCs,RefineT);
        [ti1,phi1,gi1,hi1] = make_data(dualsol{dlevel}.one,ttspan{dlevel},...
                               dual_app_K,dual_data{dlevel}.one,RefineT);
        [ti2,phi2,gi2,hi2] = make_data(dualsol{dlevel}.two,ttspan{dlevel},...
                               dual_app_K,dual_data{dlevel}.two,RefineT);
        ud_full = [ud;ud_full];
        td_full = [td;td_full];
        phi1_full = [phi1_full;phi1];
        phi2_full = [phi2_full;phi2];
        
%        rangeu = evaluate([ex_data(:,1),ex_data(:,2)],td);
%        ulevel(dlevel+1) = rangeu(end);
        ulevel(dlevel+1) = min(ud(:,1));
        ulevel= ulevel(:);
        
        [MM,NN] = size(phi1);
        if plots == 1
            figure;
            plot(phi1(1,2:NN-1)-phi0.one); 
            figure;
            plot(phi2(1,2:NN-1)-phi0.two); 
        end
    %    phi0.one = phi1(MM,2:NN-1);   %set initial condition to soln at time ts
    %    phi0.two = phi2(MM,2:NN-1);
         phi0.one = 0*x(2:end-1); disp('Initialize phi0');
         phi0.two = 0*x(2:end-1);
        
        [ux,ut] = gradient(ud,x,t);
        [phi1x,phi1t] = gradient(phi1,x,t);
        [phi2x,phi2t] = gradient(phi2,x,t);
        ux_full = [ux_full;ux];
        ut_full = [ut_full;ut];
        phi1x_full = [phi1x_full;phi1x];
        phi2x_full = [phi2x_full;phi2x];
        phi1t_full = [phi1t_full;phi1t];
        phi2t_full = [phi2t_full;phi2t];
        U{dlevel} = ud;
        PHI1{dlevel} = phi1;
        PHI2{dlevel} = phi2;
        UX{dlevel} = ux;
        UT{dlevel} = ut;
        PHI1X{dlevel} = phi1x;
        PHI1{dlevel} = phi1;
        PHI2X{dlevel} = phi2x;
        PHI2{dlevel} = phi2;

        if plots ==1
            figure;
            subplot(1,2,1);imagesc(phi1_full);
            subplot(1,2,2);imagesc(phi1x_full);
            figure;
            subplot(1,2,1);imagesc(phi2_full);
            subplot(1,2,2);imagesc(phi2x_full);
        end
        
        %% matrix system build
%        for region = 1:dlevel
         for region = dlevel
            [bigmask,smallmask] = make_mask(ud,ulevel(region:region+1));
            bmask{iter,region} =bigmask;
            Lam = (ud - ulevel(region))./(ulevel(region+1)-ulevel(region));
            M(2*dlevel-1,2*region-1) = ...
                integrate(bigmask.*Lam.*(ux-1).*flipud(phi1x),x,td);
            M(2*dlevel ,2*region-1)= ...
                integrate(bigmask.*Lam.*(ux-1).*flipud(phi2x),x,td);
            M(2*dlevel-1,2*region) = ...
                integrate(bigmask.*Lam.*ut.*flipud(phi1),x,td);
            M(2*dlevel  ,2*region)=...
                integrate(bigmask.*Lam.*ut.*flipud(phi2),x,td);
            %      area(dlevel,region) = integrate(bigmask,x,td);
        end 
        dlevel = dlevel-1;
%        keyboard

    end
    M
    level = 1;
    while (level<=iter)
        
        % check this carefully
        disp('check this carefully');
        b(2*level-1)= trapz(ttt_common{level},ddh{level}.*flipud(data1{level}(:,2)));
        b(2*level)  = trapz(ttt_common{level},ddg{level}.*flipud(data2{level}(:,3)));
        level = level+1;
    end
    % b(1:2:end) = cumsum(b(1:2:end));
    % b(2:2:end) = cumsum(b(2:2:end));
    dCK = M\b'
    % keyboard 
    dK = 1 * dCK(1:2:end); 
    dC = -1* dCK(2:2:end);
    disp('No update in diag_sys');
    dK = 0*dK;
%    dC = 0*dC;
    
    Clevel = evaluate(app_C , ulevel);
    Klevel = evaluate(app_K , ulevel);
    %modified next two line
    %fprintf('\n no update \n');   
    %    Updated_C = Clevel;
    %    Updated_K = Klevel;
    fprintf('\n Update made \n');   
    Updated_C = Clevel + relaxC.* [0;dC];
    Updated_K = Klevel + relaxK.* [0;dK];
    
    MC = max(Updated_C);
    mC = min(Updated_C);
    MK = max(Updated_K);
    mK = min(Updated_K);
    
    Updated_C = min(Cmax,Updated_C(1:end));
    Updated_C = max(Cmin,Updated_C(1:end));
    Updated_K = min(Kmax,Updated_K(1:end));
    Updated_K = max(Kmin,Updated_K(1:end));
    minu = min([gd; minu])*1.01; 
    disp('The current estimates are');
    app_C = [ [ulevel , Updated_C ]     ;
        minu  , Updated_C(end)] 
    app_K = [ [ulevel , Updated_K ]     ;
        minu  , Updated_K(end)] 
    
    %%%%%%%%%%%%%%%%%%%%
    
    if max(app_C(:,2))==Cmax | min(app_C(:,2)) == Cmin
        fprintf('Warning: The coefficient estimate C %4.1f,%4.1f is out of range [%g %g]. \n',mC,MC,Cmin,Kmax);
        fprintf('         Coefficent limited in range. The results may be inaccurate. \n');
        flagC = flagC + 1;
    end
    
    if max(app_K(:,2))==Kmax | min(app_K(:,2)) == Kmin
        fprintf('Warning: The coefficient estimate K %4.1f %4.1f is out of range [%g %g]. \n',mK,MK,Cmin,Kmax);
        fprintf('         Coefficent limited in range. The results may be inaccurate. \n');
        flagK = flagK + 1;
    end
    
    if flagC == Maxit | flagK == Maxit | flagIt == Maxit
        fprintf('         Iterated too many times (%g) or hit max or min of C %g and K  %g times. \n',flagIt,flagC,flagK);
        error('Breaking.... ' );
    end
    
    %    fprintf('Current coefficent values \n');
    %    app_C
    %    app_K
    
    CCC{iter} = app_C;
    KKK{iter} = app_K;
    M_full{iter} = M;
    b_full{iter} = b;
%    area_full{iter} = area;
    
    iter = iter+1;
end

if plots == 1;
    
    figure;
    M = length(CCC);
    N = ceil(M/2);
    if plots == 1;
        for j = 1:M 
            subplot(N,2,j);
            [c1,c2] = size(CCC{j});
            cnode = CCC{j}(1:c1-1,1);
            Cval = feval(Ctrue,cnode);
            plot(cnode,Cval,cnode,CCC{j}(1:c1-1,2),'k:');
            title(sprintf('Iteration %g',j));legend('C','my C',2);
        end
        figure;
        for j = 1:M 
            subplot(N,2,j);
            [k1,k2] = size(KKK{j});
            knode = KKK{j}(1:k1-1,1);
            Kval = feval(Ktrue,knode);
            plot(knode,Kval,knode,KKK{j}(1:k1-1,2),'k:');
            title(sprintf('Iteration %g',j));legend('K','my K',2);
        end
        xlabel(sprintf('X = (%g,%g,%g) T=(%g,%g,%g,%g) errors=(%g,%g,%g,%g)', ...
            Xmax,Xstep,Win_x,Tmax,Tstep,RefineT,Win_t,G_eps,H_eps,Tol));
        figure;
        subplot(2,1,1);
        plot(t,gtrue,'LineWidth',3);title('G(t)     (u @ x=0)');
        subplot(2,1,2);
        plot(t,htrue);title('H(t)=(K(u)*(u_x - 1)@x=1)');legend('K(um)','K(u)',4);1 
        figure;
        subplot(2,2,1);
        meshc(te,XXX,utrue');title('True Soln');xlabel('Time');ylabel('Space');
        subplot(2,2,2);
        meshc(td_full,XXX,ud_full');title('Computed U soln');xlabel('Time');ylabel('Space');
        subplot(2,2,3);
        meshc(td_full,XXX,flipud(phi1_full)');title('\Phi_1 dual soln');xlabel('Time'),ylabel('Space'); 
        subplot(2,2,4);
        meshc(td_full,XXX,flipud(phi2_full)');title('\Phi_2 dual soln');xlabel('Time'),ylabel('Space');
    end
    
end


toc

